==> lilypond-2.23.5.tar.gz: lilypond-2.23.5/lily/pdf-scheme.cc <==
/*
  This file is part of LilyPond, the GNU music typesetter.

  Copyright (C) 2011--2021 Reinhold Kainhofer <reinhold@kainhofer.com>

  LilyPond is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LilyPond is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "international.hh"
#include "memory.hh"
#include "warn.hh"
#include "lily-guile.hh"

#include <glib.h>

LY_DEFINE (ly_encode_string_for_pdf, "ly:encode-string-for-pdf",
           1, 0, 0, (SCM str),
           "Encode @var{str} as either Latin-1 (which is a subset of"
           " PDFDocEncoding) or, if that's not possible, as full UTF-16BE"
           " with a leading Byte Order Mark (BOM).")
{
  LY_ASSERT_TYPE (scm_is_string, str, 1);

  char *g = nullptr;
  gsize bytes_written = 0;

  {
    constexpr char charset[] = "UTF-8"; // Input is ALWAYS UTF-8!
    auto p (ly_scm2str0 (str));

    g = g_convert (p.get (), -1, "ASCII", charset, 0, &bytes_written, 0);
    // We do not try ISO-8859-1 since a number of versions of Ghostscript
    // versions fail to convert it to PDF.  ASCII and UTF-16BE work better with
    // recent versions of Ghostscript.

    if (!g) // ASCII not possible; resort to UTF-16BE
      {
        GError *e = NULL;
        char *g_without_BOM = g_convert (p.get (), -1, "UTF-16BE", charset, 0,
                                         &bytes_written, &e);
        if (e != NULL)
          {
            warning (_f ("Conversion of string `%s' to UTF-16be failed: %s",
                         p.get (), e->message));
            g_error_free (e);
          }
        /* UTF-16BE allows/recommends a byte-order-mark (BOM) of two bytes
         * \xFE\xFF at the begin of the string. The pdfmark specification
         * requires it and depends on it to distinguish PdfDocEncoding from
         * UTF-16BE. As g_convert does not automatically prepend this BOM
         * for UTF-16BE (only for UTF-16, which uses lower endian by default,
         * though), we have to prepend it manually. */
        if (g_without_BOM) // conversion to UTF-16be might have failed (shouldn't!)
          {
            g = static_cast<char *> (g_malloc (bytes_written + 3));
            char const *BOM = "\xFE\xFF";
            strcpy (g, BOM);
            memcpy (&g[2], g_without_BOM, bytes_written + 1); // Copy string + \0
            g_free (g_without_BOM);
            bytes_written += 2;
          }
      }
  }

  /* Convert back to SCM object and return it */
  if (g)
    {
      /*
       * Return the raw byte representation of the UTF-16BE encoded string,
       * in a locale independent way.
       */
      SCM string = scm_from_latin1_stringn (g, bytes_written);
      g_free (g);
      return string;
    }
  else
    return str;
}

==> lilypond-2.23.6.tar.gz: lilypond-2.23.6/lily/pdf-scheme.cc <==
/*
  This file is part of LilyPond, the GNU music typesetter.

  Copyright (C) 2011--2022 Reinhold Kainhofer <reinhold@kainhofer.com>

  LilyPond is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LilyPond is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "international.hh"
#include "memory.hh"
#include "warn.hh"
#include "lily-guile.hh"

#include <glib.h>

LY_DEFINE (ly_encode_string_for_pdf, "ly:encode-string-for-pdf",
           1, 0, 0, (SCM str),
           R"(
Encode @var{str} as either Latin-1 (which is a subset of PDFDocEncoding) or, if
that's not possible, as full UTF-16BE with a leading Byte Order Mark (BOM).
           )")
{
  LY_ASSERT_TYPE (scm_is_string, str, 1);

  char *g = nullptr;
  gsize bytes_written = 0;

  {
    constexpr char charset[] = "UTF-8"; // Input is ALWAYS UTF-8!
    auto p (ly_scm2str0 (str));

    g = g_convert (p.get (), -1, "ASCII", charset, 0, &bytes_written, 0);
    // We do not try ISO-8859-1 since a number of versions of Ghostscript
    // versions fail to convert it to PDF.  ASCII and UTF-16BE work better with
    // recent versions of Ghostscript.

    if (!g) // ASCII not possible; resort to UTF-16BE
      {
        GError *e = NULL;
        char *g_without_BOM = g_convert (p.get (), -1, "UTF-16BE", charset, 0,
                                         &bytes_written, &e);
        if (e != NULL)
          {
            warning (_f ("Conversion of string `%s' to UTF-16be failed: %s",
                         p.get (), e->message));
            g_error_free (e);
          }
        /* UTF-16BE allows/recommends a byte-order-mark (BOM) of two bytes
         * \xFE\xFF at the begin of the string. The pdfmark specification
         * requires it and depends on it to distinguish PdfDocEncoding from
         * UTF-16BE. As g_convert does not automatically prepend this BOM
         * for UTF-16BE (only for UTF-16, which uses lower endian by default,
         * though), we have to prepend it manually. */
        if (g_without_BOM) // conversion to UTF-16be might have failed (shouldn't!)
          {
            g = static_cast<char *> (g_malloc (bytes_written + 3));
            char const *BOM = "\xFE\xFF";
            strcpy (g, BOM);
            memcpy (&g[2], g_without_BOM, bytes_written + 1); // Copy string + \0
            g_free (g_without_BOM);
            bytes_written += 2;
          }
      }
  }

  /* Convert back to SCM object and return it */
  if (g)
    {
      /*
       * Return the raw byte representation of the UTF-16BE encoded string,
       * in a locale independent way.
       */
      SCM string = scm_from_latin1_stringn (g, bytes_written);
      g_free (g);
      return string;
    }
  else
    return str;
}
